/// Low-level Gemini server.
library gemini_server;

import 'dart:async';
import 'dart:io';
import 'package:gemini_protocol/gemini_protocol.dart';

export 'package:gemini_protocol/gemini_protocol.dart';

/// Callback that accepts a [Request] and returns a [Response].
typedef GeminiServerCallback = FutureOr<Response> Function(Request request);

/// A server that delivers content using the Gemini protocol.
///
/// When a request comes in, a [GeminiServerCallback] is executed.
/// Note: [GeminiServer] provides a low-level API. This means that
/// there are no such features as routing, middlewares, exceptions handling, etc.
class GeminiServer {
  final GeminiServerCallback _callback;
  final SecureServerSocket _serverSocket;

  GeminiServer._(this._callback, this._serverSocket) {
    _serverSocket.listen(_handler);
  }

  /// Listens to requests on a given address and port.
  /// Every time a new request is sent, [callback] is executed,
  /// the result of which is sent to the client.
  ///
  /// When the returned future completes, the server socket is bound
  /// to the given [address] and [port] and has started listening on it.
  static Future<GeminiServer> bind({
    required InternetAddress address,
    int port = 1965,
    SecurityContext? context,
    required GeminiServerCallback callback,
  }) async {
    final socket = await SecureServerSocket.bind(address, port, context);
    return GeminiServer._(callback, socket);
  }

  /// Closes this server.
  /// Once closed, the server cannot accept new requests.
  ///
  /// The returned future completes when the server
  /// is fully closed and is no longer bound.
  Future<void> close() async {
    await _serverSocket.close();
  }

  /// Handles incoming socket.
  ///
  /// Decodes the [Request], executes the callback and sends encoded [Response]
  Future<void> _handler(SecureSocket socket) async {
    final byteStream = socket.cast<List<int>>();

    late Request request;
    try {
      request = await Request.fromByteStream(byteStream);
    } on StateError catch (_) {
      return;
    }

    final response = await _callback(request);
    socket.add(response.encode());
    await socket.close();
  }
}
